from types import TracebackType
from typing import TYPE_CHECKING, Optional

from layer.common.spark import get_spark_session
from layer.dataset_build import DatasetBuild
from layer.train import Train


if TYPE_CHECKING:
    import pyspark


class Context:
    """
    Provides access to variables within the pipeline execution.

    An instance of this class is automatically passed in as an argument to the user-defined functions.
    It must always be listed as the first argument in the user-defined functions.

    This class should not be initialized by end-users.
    """

    def __init__(
        self,
        train: Optional[Train] = None,
        dataset_build: Optional[DatasetBuild] = None,
    ) -> None:
        self._spark_session: Optional["pyspark.sql.SparkSession"] = None
        self._train: Optional[Train] = train
        self._dataset_build: Optional[DatasetBuild] = dataset_build

    def spark(self) -> "pyspark.sql.SparkSession":
        """
        Retrieves the active Spark session if the model or feature is built with a Spark-specific fabric.

        :return: The active Spark session, if any are present.

        .. code-block:: python

            # Get Spark session from Layer Context.
            spark = context.spark()
        """
        if self._spark_session is None:
            self._spark_session = get_spark_session()
        return self._spark_session

    def train(self) -> Optional[Train]:
        """
        Retrieves the active Layer train object.

        :return: Represents the current train of the model, passed by Layer when the training of the model starts.

        .. code-block:: python

            # Get train object from Layer Context.
            train = context.train()
        """
        return self._train

    def dataset_build(self) -> Optional[DatasetBuild]:
        """
        Retrieves the active Layer dataset build object.

        :return: Represents the current dataset build of the dataset, passed by Layer when building of the dataset starts.
        """
        return self._dataset_build

    def with_train(self, train: Optional[Train]) -> None:
        self._train = train

    def with_dataset_build(self, dataset_build: Optional[DatasetBuild]) -> None:
        self._dataset_build = dataset_build

    def close(self) -> None:
        if self._spark_session is not None:
            self._spark_session.stop()
        self._spark_session = None

    def __enter__(self) -> "Context":
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()
