import uuid
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urlsplit

import pandas
import pandas as pd
import pyarrow
from pyarrow import flight as fl
from pyarrow.lib import ArrowKeyError

from layer.grpc_utils import create_grpc_ssl_config

from .api.service.dataset.dataset_api_pb2 import Command
from .api.value.ticket_pb2 import PartitionTicket


_CLUSTER_ID_HEADER = "cluster-id"


class DatasetClientError(Exception):
    pass


def _dataset_exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:
    def inner_function(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except ArrowKeyError as e:  # pytype: disable=mro-error
            not_found_message = str(e).replace(
                "gRPC returned not found error, with message: ",
                "",
            )
            raise DatasetClientError(not_found_message)
        except Exception as e:
            error = "Query failed"
            if args and isinstance(args[0], DatasetClient) and args[0].cluster_id:
                error += f", cluster id: {args[0].cluster_id}"
            raise DatasetClientError(error) from e

    return inner_function


@dataclass(frozen=True)
class PartitionMetadata:
    location: str
    format: int


class Partition:
    def __init__(self, reader: Union[fl.FlightStreamReader, pd.DataFrame]):
        self._reader = reader

    def to_pandas(self) -> pandas.DataFrame:
        if isinstance(self._reader, pandas.DataFrame):
            return self._reader
        return self._reader.read_pandas()


class _FlightCallMetadataMiddleware(fl.ClientMiddleware):
    def __init__(self, call_metadata: List[Tuple[str, str]], client: "DatasetClient"):
        super().__init__()
        self._call_metadata = call_metadata
        self._client = client

    def sending_headers(self) -> Any:
        headers = {i[0]: i[1] for i in self._call_metadata}
        headers["x-request-id"] = str(uuid.uuid4())
        return headers

    def received_headers(self, headers: Dict[str, str]) -> Any:
        if _CLUSTER_ID_HEADER in headers:
            self._client.cluster_id = headers[_CLUSTER_ID_HEADER][0]


class _FlightCallMetadataMiddlewareFactory(fl.ClientMiddlewareFactory):
    def __init__(self, call_metadata: List[Tuple[str, str]], client: "DatasetClient"):
        super().__init__()
        self._call_metadata = call_metadata
        self._client = client

    def start_call(self, info: fl.CallInfo) -> Any:
        return _FlightCallMetadataMiddleware(self._call_metadata, self._client)


class DatasetClient:
    def __init__(self, address_and_port: str, access_token: str) -> None:
        self._endpoint = f"grpc+tls://{address_and_port}"
        self._ssl_config = create_grpc_ssl_config(
            address_and_port, do_verify_ssl=True, do_force_cadata_load=True
        )
        self._middleware = [
            _FlightCallMetadataMiddlewareFactory(
                [("authorization", f"Bearer {access_token}")], self
            )
        ]
        self._flight = fl.FlightClient(
            self._endpoint,
            override_hostname=f"dataset.{address_and_port.partition(':')[0]}",
            tls_root_certs=self._ssl_config.cadata,
            middleware=self._middleware,
        )
        self._cluster_id: Optional[str] = None

    @property
    def cluster_id(self) -> Optional[str]:
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id: Optional[str]) -> None:
        self._cluster_id = cluster_id

    @_dataset_exception_handler
    def health_check(self) -> str:
        buf = pyarrow.allocate_buffer(0)
        action = fl.Action("HealthCheck", buf)
        result = next(self._flight.do_action(action))
        return result.body.to_pybytes().decode("utf-8")

    @_dataset_exception_handler
    def get_partitions_metadata(self, command: Command) -> List[PartitionMetadata]:
        descriptor = fl.FlightDescriptor.for_command(command.SerializeToString())
        flight_info = self._flight.get_flight_info(descriptor)
        partitions_metadata = []
        for endpoint in flight_info.endpoints:
            ticket = PartitionTicket()
            ticket.ParseFromString(endpoint.ticket.ticket)
            for location in endpoint.locations:
                partitions_metadata.append(
                    PartitionMetadata(
                        location=location.uri.decode("utf-8"), format=ticket.format
                    )
                )
        return partitions_metadata

    @_dataset_exception_handler
    def fetch_partition(self, partition_metadata: PartitionMetadata) -> Partition:
        if partition_metadata.format == PartitionTicket.Format.FORMAT_PARQUET:
            return Partition(pd.read_parquet(partition_metadata.location))

        location = urlsplit(partition_metadata.location).hostname
        partition_client = fl.FlightClient(
            self._endpoint,
            override_hostname=location,
            tls_root_certs=self._ssl_config.cadata,
            middleware=self._middleware,
        )
        reader = partition_client.do_get(fl.Ticket(b""))
        return Partition(reader)

    @_dataset_exception_handler
    def get_dataset_writer(self, command: Command, schema: Any) -> Any:
        descriptor = fl.FlightDescriptor.for_command(command.SerializeToString())
        return self._flight.do_put(descriptor, schema)
