import io
from pathlib import Path
from traceback import FrameSummary
from typing import Any, List, Optional, Set, Type

import jsonschema
import yaml
from yarl import URL

from layer.api.ids_pb2 import RunId

from .status_report import ExecutionStatusReport


RICH_ERROR_COLOR = "rgb(251,147,60)"


class RuntimeMemoryException(Exception):
    def __init__(self, err_details: str):
        super().__init__(f"Memory error: {err_details}")


class FunctionParameterException(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class SparkRuntimeException(Exception):
    pass


class ConfigError(Exception):
    pass


class MissingConfigurationError(ConfigError):
    def __init__(self, path: Path):
        self._path: Path = path
        super().__init__(f"Missing configuration file {path.resolve()}")

    @property
    def path(self) -> Path:
        return self._path


class InvalidConfigurationError(ConfigError):
    def __init__(self, path: Path):
        self._path: Path = path
        super().__init__(f"Invalid configuration file {path.resolve()}")

    @property
    def path(self) -> Path:
        return self._path


class AuthException(Exception):
    pass


class LayerClientException(Exception):
    pass


class LayerClientTimeoutException(LayerClientException):
    pass


class LayerClientResourceNotFoundException(LayerClientException):
    pass


class LayerClientResourceAlreadyExistsException(LayerClientException):
    pass


class LayerResourceExhaustedException(LayerClientException):
    pass


class LayerClientServiceUnavailableException(LayerClientException):
    pass


class LayerClientUnknownIntegrationProviderException(LayerClientException):
    pass


class LayerClientUnsupportedIntegrationProviderException(LayerClientException):
    pass


class ProjectException(Exception):
    pass


class ProjectBaseException(Exception):
    def __init__(self, error_msg: str = "", suggestion: str = ""):
        self._error_msg = error_msg
        self._suggestion = suggestion
        super().__init__(self._format_message())

    @property
    def error_msg(self) -> str:
        return self._error_msg

    @property
    def error_msg_rich(self) -> str:
        return self.error_msg

    @property
    def suggestion(self) -> str:
        return self._suggestion

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def _format_message(self) -> str:
        return f"Error: {self._error_msg},\nSuggestion: {self._suggestion}"


class ProjectExecutionException(ProjectBaseException):
    def __init__(self, run_id: RunId, err_msg: str, suggestion: str):
        super().__init__(err_msg, suggestion)
        self._run_id = run_id

    def run_id(self) -> RunId:
        return self._run_id


class ProjectModelExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, train_id: str, report: ExecutionStatusReport):
        self._report = report
        self._train_id = train_id
        super().__init__(
            run_id,
            self.__format_message(),
            f'Run `layer.logs("{run_id.value}")` for additional info',
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f"Train [bold {RICH_ERROR_COLOR}]{self._train_id}[/bold {RICH_ERROR_COLOR}] failed with "
            + f'"{self._report.message}"'
            + "\nCause: "
            + self._report.cause
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return (
            f"Train {self._train_id} failed with "
            + f'"{self._report.message}"'
            + "\nCause: "
            + self._report.cause
        )

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectHPTExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, tuning_id: str, report: ExecutionStatusReport):
        self._tuning_id = tuning_id
        self._report = report
        super().__init__(
            run_id,
            self.__format_message(),
            f'Run `layer.logs("{run_id.value}")` for additional info',
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f"HPT [bold {RICH_ERROR_COLOR}]{self._tuning_id}[/bold {RICH_ERROR_COLOR}] failed with "
            + f'"{self._report.message}"'
            + "\nCause: "
            + self._report.cause
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return (
            f"HPT {self._tuning_id} failed with "
            + f'"{self._report.message}"'
            + "\nCause: "
            + self._report.cause
        )

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectFeatureBuildExecutionException(ProjectExecutionException):
    def __init__(
        self, run_id: RunId, featureset_id: str, report: ExecutionStatusReport
    ):
        self._featureset_id = featureset_id
        self._report = report
        super().__init__(
            run_id,
            self.__format_message(),
            f'Run `layer.logs("{run_id.value}")` for additional info',
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f'Featureset "[bold {RICH_ERROR_COLOR}]{self._featureset_id}[/bold {RICH_ERROR_COLOR}]" build failed with '
            + f'"{self._report.message}"'
            + ("\nCause: " + self._report.cause if self._report.cause != "" else "")
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return (
            f'Featureset "{self._featureset_id}" build failed with '
            + f'"{self._report.message}"'
            + ("\nCause: " + self._report.cause if self._report.cause != "" else "")
        )

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectDatasetBuildExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, dataset_id: str, report: ExecutionStatusReport):
        self._dataset_id = dataset_id
        self._report = report
        super().__init__(
            run_id,
            self.__format_message(),
            f'Run `layer.logs("{run_id.value}")` for additional info',
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f'Dataset "[bold {RICH_ERROR_COLOR}]{self._dataset_id}[/bold {RICH_ERROR_COLOR}]" build failed with '
            + f'"{self._report.message}"'
            + ("\nCause: " + self._report.cause if self._report.cause != "" else "")
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return (
            f'Dataset "{self._dataset_id}" build failed with '
            + f'"{self._report.message}"'
            + ("\nCause: " + self._report.cause if self._report.cause != "" else "")
        )

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectInitializationException(ProjectBaseException):
    pass


class ProjectDependencyNotFoundException(ProjectInitializationException):
    pass


class LayerServiceUnavailableExceptionDuringInitialization(
    ProjectInitializationException
):
    def __init__(self, message: str):
        super().__init__(
            f"Layer service unavailable. {message}",
            "Check your internet connection.",
        )


class LayerServiceUnavailableExceptionDuringExecution(ProjectExecutionException):
    def __init__(self, run_id: RunId, message: str):
        super().__init__(
            run_id,
            f"Layer service unavailable. {message}",
            "Check your internet connection.",
        )


class NoProjectFoundException(ProjectInitializationException):
    def __init__(self, path: Path):
        self._path: Path = path
        super().__init__(
            f'Path "{path.resolve()}" not within a Layer project. Missing ".layer/project.yml" file.',
            "Please initialize a new project by running `layer init`.",
        )


class UnknownIntegrationException(ProjectInitializationException):
    def __init__(self, name: str):
        super().__init__(
            f"No integration with name {name}.",
            "Please, check if the name is correct or look how to add the integration with `layer integration --help`.",
        )


class UnknownIntegrationProviderException(ProjectInitializationException):
    def __init__(self, name: str):
        super().__init__(
            f"Unknown integration type for {name}.",
            "Please, check supported integration providers with `layer integration --help`.",
        )


class UnsupportedIntegrationProviderException(ProjectInitializationException):
    def __init__(self, name: str):
        super().__init__(
            f"Integration provider for integration {name} is not supported.",
            "Please, check supported integration providers with `layer integration --help`.",
        )


class InvalidEntityTypeYamlDefinitionError(ProjectInitializationException):
    def __init__(self, path: Path, entity_type: str):
        super().__init__(
            f"Unsupported Layer entity of type '{entity_type}' used in {path}",
            "Please use one of 'raw_dataset', 'derived_dataset', 'featureset' or 'model'",
        )


class ProjectInvalidYamlException(ProjectInitializationException):
    def __init__(self, yaml_path: Path, exc: yaml.YAMLError):
        super().__init__(
            f'Unable to parse "{yaml_path}":\n{exc}',
            "Please, check your YAML syntax.",
        )


class UserNotLoggedInException(ProjectInitializationException):
    def __init__(self) -> None:
        super().__init__("User not logged in.", "Run `layer.login()`")


class UserConfigurationError(ProjectInitializationException):
    def __init__(self, path: Path) -> None:
        super().__init__(
            f"User configuration error in {path.resolve()}",
            "Please make sure that you have the latest version and try to login again.",
        )


class UserNotInAnOrganizationError(ProjectInitializationException):
    def __init__(self, url: URL) -> None:
        super().__init__(
            "You are not yet member of an organization!",
            f"You can either create an organization at {url} or be invited to an already existing one.",
        )


class UserAccessTokenExpiredError(ProjectInitializationException):
    def __init__(
        self,
    ) -> None:
        super().__init__("Your session has expired.", "Please, try to login again.")


class ProjectRunnerError(Exception):
    def __init__(self, message: str = "", run_id: Optional[RunId] = None):
        super().__init__(message)
        self._run_id = run_id

    def run_id(self) -> Optional[RunId]:
        return self._run_id


class FunctionDefinitionException(ProjectInitializationException):
    def __init__(
        self,
        param_name: str,
        func_name: str,
        script_path: Path,
        valid_entities: Set[str],
        package_name: str,
    ) -> None:
        msg = f'Parameter "{param_name}" of function "{func_name}" in "{script_path.resolve()}" is not annotated with one of {valid_entities} {package_name} package entities.'
        super().__init__(
            msg, "Please, look at the docs for expected function definition."
        )


class MissingFunctionException(ProjectInitializationException):
    def __init__(
        self,
        func_name: str,
        script_path: Path,
    ) -> None:
        msg = f'Function "{func_name}" not defined in {script_path.resolve()}'
        super().__init__(
            msg, "Please, look at the docs for expected function definition."
        )


class ProjectRunTerminatedError(ProjectRunnerError):
    pass


class IntegrationException(Exception):
    pass


class UnexpectedModelTypeException(Exception):
    def __init__(self, model_object_type: Type[Any]):
        super().__init__(
            f"train_model function returned an unsupported model object type '{model_object_type}'"
        )


class ProjectCircularDependenciesException(ProjectInitializationException):
    def __init__(self, stringified_paths: List[str]):
        super().__init__(
            self._err_msg(stringified_paths),
            "Please, check your project dependencies for errors",
        )
        self._paths: List[str] = stringified_paths

    @property
    def stringified_cycle_paths(self) -> List[str]:
        return self._paths

    @staticmethod
    def _err_msg(stringified_paths: List[str]) -> str:
        paths_as_str = "\n".join(stringified_paths)
        return f"Detected circular dependencies:\n{paths_as_str}"


class ProjectDefinitionError(ProjectInitializationException):
    pass


class ProjectFieldError(ProjectDefinitionError):
    def __init__(self, message: str, field_chain: List[str], field: str):
        self.field_chain = field_chain
        self.field = field
        self.message = message
        super().__init__(self.__str__(), "Please, fix the field value and try again.")

    def _get_field_chain(self) -> str:
        extended = self.field_chain.copy()
        extended.append(self.field)
        return ".".join(extended)

    def __str__(self) -> str:
        return f"{self.message}.\nProperty: {self._get_field_chain()}"


class ProjectSchemaValidationError(ProjectDefinitionError):
    def __init__(
        self,
        exc: jsonschema.exceptions.ValidationError,
        file_path: Path,
        parent_field_chain: List[str],
    ):
        self.file_path = file_path
        self.parent_field_chain = parent_field_chain
        self.exc = exc
        super().__init__(
            self.__str__(),
            "Please check the docs for a correct project YAML definition.",
        )

    def _get_field_chain(self) -> str:
        chain = ["ROOT"]
        chain.extend(self.parent_field_chain)
        chain.extend(self.exc.absolute_path)
        return ".".join([str(chain_elem) for chain_elem in chain])

    def __str__(self) -> str:
        if isinstance(self.exc.instance, dict):
            stream = io.StringIO()
            yaml.safe_dump(self.exc.instance, stream=stream)
            invalid_definition = "\n" + stream.getvalue()
        else:
            invalid_definition = self.exc.instance
        return (
            f"{self.exc.message}.\n"
            f'File: "{self.file_path.resolve()}"\n'
            f'Parent property: "{self._get_field_chain()}".\n'
            f"Invalid definition: {invalid_definition}"
        )


class NonDataFrameTypeException(Exception):
    def __init__(self, object_type: Type[Any]):
        super().__init__(
            f"function returned an unsupported object type '{object_type}'"
        )


class MissingColumnsInDataframeException(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Dataframe is required to have at least two columns and must contain the primary key and the value."
        )
