import ast
from ast import Attribute, Call, FunctionDef, Import, ImportFrom, Module, Name
from pathlib import Path
from typing import Optional, Set

from ..client import BasePythonFeature, PythonDataset
from ..exceptions.exceptions import (
    FunctionDefinitionException,
    FunctionParameterException,
    MissingFunctionException,
)
from ..projects.project import Project


LAYER_PACKAGE_NAME = "layer"
VALID_TRAIN_MODEL_FUNCTION_ENTITIES = {
    "Train",
    "Context",
    "Dataset",
    "Featureset",
    "Model",
}
VALID_BUILD_FEATURE_FUNCTION_ENTITIES = {"Context", "Dataset", "Featureset", "Model"}
VALID_BUILD_DERIVED_DATASET_FUNCTION_ENTITIES = {
    "Context",
    "Dataset",
    "Featureset",
    "Model",
}
FUNCTION_CONTEXT_ENTITY = "Context"


class CodeParser:
    def __init__(self, project: Project):
        self._project = project

    def validate(self) -> None:
        self._validate_models()
        self._validate_featuresets()
        self._validate_derived_datasets()

    def _validate_models(self) -> None:
        for model in self._project.models:
            source_dir = model.local_path.parent
            entrypoint_path = source_dir.resolve() / model.training.entrypoint
            FunctionValidator(
                entrypoint_path, "train_model", VALID_TRAIN_MODEL_FUNCTION_ENTITIES
            ).validate()

    def _validate_featuresets(self) -> None:
        for featureset in self._project.featuresets:
            for feature in featureset.features:
                if isinstance(feature, BasePythonFeature):
                    entrypoint_path = feature.source_path
                    FunctionValidator(
                        entrypoint_path,
                        "build_feature",
                        VALID_BUILD_FEATURE_FUNCTION_ENTITIES,
                    ).validate()

    def _validate_derived_datasets(self) -> None:
        for dataset in self._project.derived_datasets:
            if isinstance(dataset, PythonDataset):
                entrypoint_path = dataset.entrypoint_path
                FunctionValidator(
                    entrypoint_path,
                    "build_dataset",
                    VALID_BUILD_DERIVED_DATASET_FUNCTION_ENTITIES,
                ).validate()


class FunctionValidator:
    """
    Checks that the given function defined in the path script is defined and with valid parameter annotations
    :param path: Path to the python script
    :param func_name: Name of the function to check
    :param valid_entities: What annotations are allowed for this function's params
    """

    def __init__(self, path: Path, func_name: str, valid_entities: Set[str]) -> None:
        self._path = path
        self._func_name = func_name
        self._valid_entities = valid_entities
        with open(path, mode="r") as file:
            content = file.read()
        self._tree = ast.parse(content)

    @staticmethod
    def _get_layer_package_name(ast_node: Module) -> Optional[str]:
        """
        Infers how the layer package could be referenced in the code(including aliases), if it is imported
        """
        import_statements = [n for n in ast_node.body if isinstance(n, Import)]
        for import_statement in import_statements:
            for imported_name in import_statement.names:
                if imported_name.name == LAYER_PACKAGE_NAME:
                    if imported_name.asname is not None:
                        return imported_name.asname
                    else:
                        return imported_name.name
        return None

    def _validate_ast_name(
        self, parameter_name: str, ast_annotation: Name, must_be_context: bool
    ) -> None:
        try:
            entity = self._check_layer_entity_import(
                self._tree, parameter_name, ast_annotation.id, self._valid_entities
            )
        except Exception:
            raise FunctionDefinitionException(
                parameter_name,
                self._func_name,
                self._path,
                self._valid_entities,
                LAYER_PACKAGE_NAME,
            )
        if must_be_context and entity != FUNCTION_CONTEXT_ENTITY:
            raise FunctionParameterException(
                "Context parameter is required as a first parameter"
            )

    def _validate_ast_attribute(
        self, parameter_name: str, ast_annotation: Attribute, must_be_context: bool
    ) -> None:
        if isinstance(ast_annotation.value, Name):
            coming_from = ast_annotation.value.id
            entity_name = ast_annotation.attr
            layer_package = self._get_layer_package_name(
                self._tree
            )  # The alias for the layer package
            if coming_from != layer_package:
                raise FunctionDefinitionException(
                    parameter_name,
                    self._func_name,
                    self._path,
                    self._valid_entities,
                    LAYER_PACKAGE_NAME,
                )
            if must_be_context and entity_name != FUNCTION_CONTEXT_ENTITY:
                raise FunctionParameterException(
                    "Context parameter is required as a first parameter"
                )
            if entity_name not in self._valid_entities:
                raise FunctionDefinitionException(
                    parameter_name,
                    self._func_name,
                    self._path,
                    self._valid_entities,
                    LAYER_PACKAGE_NAME,
                )

    def validate(self) -> None:
        layer_func = self._extract_function(self._tree, self._func_name)
        if layer_func is None:
            raise MissingFunctionException(self._func_name, self._path)
        for idx in range(len(layer_func.args.args)):
            arg = layer_func.args.args[idx]
            parameter_name = arg.arg
            must_be_context = idx == 0
            if arg.annotation is None:
                raise FunctionDefinitionException(
                    parameter_name,
                    self._func_name,
                    self._path,
                    self._valid_entities,
                    LAYER_PACKAGE_NAME,
                )
            elif isinstance(arg.annotation, Name):
                self._validate_ast_name(parameter_name, arg.annotation, must_be_context)
            elif isinstance(arg.annotation, Attribute):
                self._validate_ast_attribute(
                    parameter_name, arg.annotation, must_be_context
                )
            elif isinstance(arg.annotation, Call):
                if isinstance(arg.annotation.func, Name):
                    self._validate_ast_name(
                        parameter_name, arg.annotation.func, must_be_context
                    )
                elif isinstance(arg.annotation.func, Attribute):
                    self._validate_ast_attribute(
                        parameter_name, arg.annotation.func, must_be_context
                    )

    @staticmethod
    def _check_layer_entity_import(
        tree: Module, parameter_name: str, search_for: str, valid_entities: Set[str]
    ) -> str:
        """
        Check that search_for is imported from the layer package and is also among the valid annotation entities
        :param tree: The AST for a Python script
        :param parameter_name: Which parameter's annotation we are checking
        :param search_for: Annotation name we're checking
        :param valid_entities: The set of valid annotation types
        :return: which of the valid entities matches the param's annotation
        """
        import_froms = [n for n in tree.body if isinstance(n, ImportFrom)]
        for import_from in import_froms:
            if import_from.module == LAYER_PACKAGE_NAME:
                for imported_name in import_from.names:
                    import_obj = ImportName(imported_name.name, imported_name.asname)
                    if (
                        import_obj.matches_type(search_for)
                        and import_obj.name in valid_entities
                    ):
                        return imported_name.name
        raise Exception(
            f'Parameter "{parameter_name}" not annotated with one of {valid_entities} Layer entities'
        )

    @staticmethod
    def _extract_function(tree: Module, func_name: str) -> Optional[FunctionDef]:
        """
        Return the AST function definition for the given function name.
        """
        functions = [n for n in tree.body if isinstance(n, FunctionDef)]
        for function in functions:
            if function.name == func_name:
                return function
        return None


class ImportName:
    def __init__(self, name: str, alias: Optional[str]):
        """
        :param name: import name
        :param alias: alias for the imported name
        """
        self._name = name
        self._alias = alias

    @property
    def name(self) -> str:
        """
        :return: the name of the import
        """
        return self._name

    def matches_type(self, aliased_type: str) -> bool:
        """
        Check if the given type name(which might be an alias) matches the imported name
        :param aliased_type: potentially aliased type name
        :return: whether the current import corresponds to the given type
        """
        if self._alias is not None:
            return self._alias == aliased_type
        else:
            return self._name == aliased_type
