import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional

import jsonschema
import yaml

from layer.exceptions.exceptions import (
    ProjectDefinitionError,
    ProjectInvalidYamlException,
    ProjectSchemaValidationError,
)


SCHEMA_BASE_PATH = (
    Path(__file__) / ".." / ".." / "assets" / "project_schema"
).resolve()

FieldChain = Optional[List[str]]


class ProjectDefinitionParser(ABC):
    def __init__(
        self,
        path: Path,
        project_name: str,
        json_schema_file_name: str,
        payload: Optional[Dict[str, Any]] = None,
        field_chain: Optional[FieldChain] = None,
    ):
        self.__data = (
            payload.copy() if payload is not None else self._read_payload(path)
        )
        self.__schema = self._read_schema(SCHEMA_BASE_PATH / json_schema_file_name)

        self._path = path
        self._project_name = project_name
        self._field_chain: List[str] = field_chain if field_chain is not None else []

    def validate(self) -> None:
        try:
            jsonschema.validate(instance=self.__data, schema=self.__schema)
        except jsonschema.exceptions.ValidationError as e:
            raise ProjectSchemaValidationError(e, self._path, self._field_chain)
        except jsonschema.exceptions.SchemaError:
            raise ProjectDefinitionError("Internal project validation error")

    def _safe_payload(self) -> Dict[str, Any]:
        self.validate()
        return self.__data

    @abstractmethod
    def parse(self) -> Any:
        pass

    @staticmethod
    def _read_payload(path: Path) -> Dict[str, Any]:
        try:
            with open(path, "r") as f:
                return yaml.safe_load(f)
        except yaml.YAMLError as e:
            raise ProjectInvalidYamlException(path, e)

    @staticmethod
    def _read_schema(path: Path) -> Dict[str, Any]:
        with open(path) as json_file:
            return json.load(json_file)
