from pathlib import Path
from typing import Any, Dict, Optional

from ..annotation_processor import parse_entity_dependencies_from_file
from ..client import (
    DatasourceRef,
    DatasourceType,
    DerivedDataset,
    PythonDataset,
    SparkDataset,
    SQLDataset,
)
from ..exceptions.exceptions import ProjectException
from ..projects.asset import AssetPath, AssetType
from .common import FieldChain, ProjectDefinitionParser


class DerivedDatasetDefinitionParser(ProjectDefinitionParser):
    def __init__(
        self,
        payload: Dict[str, Any],
        path: Path,
        project_name: str,
        field_chain: Optional[FieldChain] = None,
    ):
        super().__init__(
            path, project_name, "derived_dataset.json", payload, field_chain
        )

    def parse(self) -> DerivedDataset:
        _data = self._safe_payload()
        materialization = _data.get("materialization", {})
        integration_name = materialization.get("integration_name", None)
        environment = _data.get("environment", None)
        entrypoint = _data.get("entrypoint", None)
        fabric = _data.get("fabric", None)
        name = _data.get("name", None)
        description = _data.get("description", "")
        if self._is_sql_dataset(entrypoint):
            entrypoint_path = self._path.parent / entrypoint
            query_text = entrypoint_path.read_text()
            asset_path = AssetPath(
                entity_name=name,
                asset_type=AssetType.DATASET,
                project_name=self._project_name,
            )
            return SQLDataset(
                asset_path=asset_path,
                description=description,
                query_text=query_text,
                datasource=DatasourceRef(
                    name=integration_name, type=DatasourceType.TABLE
                ),
            )
        else:
            if not environment:
                raise ProjectException(
                    f"No environment specified for the derived dataset {name!r}."
                )
            entrypoint_path = self._path.parent / entrypoint
            entrypoint_content = entrypoint_path.read_text()
            environment_path = self._path.parent / environment
            dependencies = parse_entity_dependencies_from_file(
                entrypoint_path.parent, entrypoint_path.name, "build_dataset"
            )

            asset_path = AssetPath(
                entity_name=name,
                asset_type=AssetType.DATASET,
                project_name=self._project_name,
            )

            return (
                SparkDataset(
                    asset_path=asset_path,
                    description=description,
                    entrypoint=entrypoint_path.name,
                    entrypoint_content=entrypoint_content,
                    entrypoint_path=entrypoint_path,
                    environment=environment_path.name,
                    environment_path=environment_path,
                    dependencies=dependencies,
                    fabric=fabric,
                    datasource=DatasourceRef(
                        name=integration_name, type=DatasourceType.TABLE
                    ),
                )
                if fabric and "spark" in fabric
                else PythonDataset(
                    asset_path=asset_path,
                    description=description,
                    entrypoint=entrypoint_path.name,
                    entrypoint_content=entrypoint_content,
                    entrypoint_path=entrypoint_path,
                    environment=environment_path.name,
                    environment_path=environment_path,
                    dependencies=dependencies,
                    fabric=fabric,
                    datasource=DatasourceRef(
                        name=integration_name, type=DatasourceType.TABLE
                    ),
                )
            )

    @staticmethod
    def _is_sql_dataset(entrypoint: Optional[str]) -> bool:
        return entrypoint is not None and entrypoint.endswith(".sql")
