from pathlib import Path
from typing import Any, Dict, List, Optional

from layer.exceptions.exceptions import ProjectException

from ..annotation_processor import parse_entity_dependencies_from_file
from ..client import (
    Aggregation,
    DatasourceRef,
    DatasourceType,
    Feature,
    Featureset,
    PythonFeature,
    SparkFeature,
    SQLFeature,
)
from ..projects.asset import AssetPath, AssetType
from .common import FieldChain, ProjectDefinitionParser


class FeaturesetDefinitionParser(ProjectDefinitionParser):
    def __init__(
        self,
        payload: Dict[str, Any],
        path: Path,
        project_name: str,
        field_chain: Optional[FieldChain] = None,
    ):
        super().__init__(path, project_name, "featureset.json", payload, field_chain)

    def parse(self) -> Featureset:
        _data = self._safe_payload()
        featureset_name = _data.get("name", None)
        features = self.parse_features(featureset_name)
        datasource = self._load_featureset_datasource(_data)
        dependencies = [feature.dependencies for _, feature in features.items()]

        asset_path = AssetPath(
            entity_name=featureset_name,
            asset_type=AssetType.DATASET,
            project_name=self._project_name,
        )

        return Featureset(
            asset_path=asset_path,
            description=_data.get("description", None),
            features=list(features.values()),
            dependencies=sum(dependencies, []),
            datasource=datasource,
        )

    def parse_features(self, featureset_name: str) -> Dict[str, Feature]:
        _data = self._safe_payload()
        features: Dict[str, Feature] = {}
        featureset_fabric = _data.get("fabric", "f-small")
        for feature_payload in _data.get("features", []):
            feature_fabric = self._load_feature_fabric(
                featureset_fabric, feature_payload
            )
            feature_payload["fabric"] = feature_fabric
            feature = self._load_feature(self._path.parent, feature_payload)
            if feature.name in features:
                raise ProjectException(
                    f"Duplicate feature {feature.name!r} found in featureset {featureset_name!r}"
                )
            features[feature.name] = feature

        return features

    def _load_feature(self, path: Path, feature_payload: Dict[str, Any]) -> Feature:
        name = feature_payload["name"]
        description = feature_payload.get("description", None)
        aggregations = self._load_aggregations(feature_payload)
        if self._is_sql_feature(feature_payload):
            if "source" in feature_payload:
                source_path = path / feature_payload["source"]
                query = source_path.read_text()
            else:
                query = feature_payload["sql"]
            return SQLFeature(
                name=name,
                description=description,
                query=query,
                aggregations=aggregations,
            )
        else:
            source_path = path / feature_payload["source"]
            if source_path.suffix not in [".py"]:
                raise ProjectException(
                    f"Unsupported source file format: {source_path.suffix}"
                )
            if "environment" not in feature_payload:
                raise ProjectException(
                    f"No environment specified for the feature {name!r}."
                )
            environment_path = path / feature_payload["environment"]
            dependencies = parse_entity_dependencies_from_file(
                source_path.parent, source_path.name, "build_feature"
            )
            fabric = feature_payload["fabric"]
            if "spark" in fabric:
                return SparkFeature(
                    name=name,
                    description=description,
                    source=source_path.name,
                    source_path=source_path,
                    environment=environment_path.name,
                    environment_path=environment_path,
                    aggregations=aggregations,
                    dependencies=dependencies,
                    fabric=fabric,
                )
            else:
                return PythonFeature(
                    name=name,
                    description=description,
                    source=source_path.name,
                    source_path=source_path,
                    environment=environment_path.name,
                    environment_path=environment_path,
                    aggregations=aggregations,
                    dependencies=dependencies,
                    fabric=fabric,
                )

    @staticmethod
    def _load_featureset_datasource(_data: Dict[str, Any]) -> DatasourceRef:
        metadata = _data.get("materialization", {})
        datasource_name = metadata.get("integration_name")
        datasource = DatasourceRef(name=datasource_name, type=DatasourceType.STORAGE)

        return datasource

    @staticmethod
    def _load_feature_fabric(
        featureset_fabric: str, feature_payload: Dict[str, Any]
    ) -> str:
        if "fabric" in feature_payload:
            feature_fabric = (
                "f-spark-small"
                if feature_payload["fabric"] == "spark"
                else feature_payload["fabric"]
            )
            return feature_fabric
        else:
            return featureset_fabric

    @staticmethod
    def _is_sql_feature(feature_payload: Dict[str, Any]) -> bool:
        if "sql" in feature_payload:
            return True
        elif "source" in feature_payload:
            return feature_payload["source"].endswith(".sql")
        return False

    @staticmethod
    def _load_aggregations(feature_payload: Dict[str, Any]) -> List[Aggregation]:
        return [
            Aggregation(method=agg.get("method"), windows=agg.get("windows"))
            for agg in feature_payload.get("aggregations", [])
        ]
