import hashlib
import itertools
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from layer.exceptions.exceptions import ProjectException
from layer.projects.asset import AssetPath, AssetType

from ..annotation_processor import parse_entity_dependencies_from_file
from ..client import (
    BayesianSearch,
    GridSearch,
    HyperparameterTuning,
    LayerClientException,
    ManualSearch,
    Model,
    Parameter,
    ParameterCategoricalRange,
    ParameterRange,
    ParameterStepRange,
    ParameterType,
    ParameterValue,
    RandomSearch,
    Train,
    TypedParameter,
)
from .common import FieldChain, ProjectDefinitionParser


class ModelDefinitionParser(ProjectDefinitionParser):
    def __init__(
        self,
        payload: Dict[str, Any],
        path: Path,
        project_name: str,
        field_chain: Optional[FieldChain] = None,
    ):
        super().__init__(path, project_name, "model.json", payload, field_chain)

    def parse(self) -> Model:
        payload = self._safe_payload()
        if "name" not in payload:
            raise ProjectException("No name specified for the model.")
        name = payload["name"]
        description = ""
        if "description" in payload:
            description = payload["description"]
        if "training" not in payload:
            raise ProjectException(f"No training specified for the model {name!r}.")
        training = self._load_train(payload["training"])
        dependencies = parse_entity_dependencies_from_file(
            self._path.parent, training.entrypoint, "train_model"
        )
        training_files_hash = hashlib.sha256()
        training_files = self._path.parent.glob("*.py")
        environment_files = []
        if training.environment and training.environment != "":
            environment_files.append(self._path.parent / training.environment)
        for training_file_path in itertools.chain(training_files, environment_files):
            with open(training_file_path, "rb") as training_file:
                training_files_bytes = training_file.read()
                training_files_hash.update(training_files_bytes)

        asset_path = AssetPath(
            entity_name=name,
            asset_type=AssetType.MODEL,
            project_name=self._project_name,
        )

        return Model(
            asset_path=asset_path,
            local_path=self._path,
            description=description,
            training=training,
            dependencies=dependencies,
            training_files_digest=training_files_hash.hexdigest(),
        )

    def _load_train(self, training_payload: Dict[str, Any]) -> Train:
        name = training_payload["name"]
        description = training_payload["description"]
        entrypoint = training_payload["entrypoint"]
        environment = training_payload["environment"]
        parameters: List[Parameter] = []
        hyperparameter_tuning: Optional[HyperparameterTuning] = None
        if "hyperparameters" in training_payload:
            hyperparameter_tuning = self._parse_hyperparameters(
                training_payload["hyperparameters"]
            )
        for param_payload in training_payload.get("parameters", []):
            parameters.append(
                Parameter(name=param_payload["name"], value=param_payload["value"])
            )
        fabric: str = training_payload.get("fabric", "")
        return Train(
            name=name,
            description=description,
            entrypoint=entrypoint,
            environment=environment,
            parameters=parameters,
            hyperparameter_tuning=hyperparameter_tuning,
            fabric=fabric,
        )

    def _parse_hyperparameters(
        self, hyper_params_payload: Dict[str, Any]
    ) -> HyperparameterTuning:
        strategy: str = hyper_params_payload["strategy"].lower().strip()
        manual_search: Optional[ManualSearch] = None
        random_search: Optional[RandomSearch] = None
        grid_search: Optional[GridSearch] = None
        bayesian_search: Optional[BayesianSearch] = None
        if strategy == "manual":
            manual_search = self._parse_manual_strategy(hyper_params_payload)
        elif strategy == "random":
            random_search = self._parse_random_strategy(hyper_params_payload)
        elif strategy == "grid":
            grid_search = self._parse_grid_strategy(hyper_params_payload)
        elif strategy == "bayesian":
            bayesian_search = self._parse_bayesian_strategy(hyper_params_payload)
        else:
            raise LayerClientException(f"Unsupported strategy type {strategy}")

        return HyperparameterTuning(
            strategy=strategy,
            max_parallel_jobs=hyper_params_payload.get("max_parallel_trials", None),
            maximize=hyper_params_payload.get("maximize", None),
            minimize=hyper_params_payload.get("minimize", None),
            early_stop=hyper_params_payload.get("early_stop", None),
            fixed_parameters=hyper_params_payload.get("fixed", {}),
            manual_search=manual_search,
            random_search=random_search,
            grid_search=grid_search,
            bayesian_search=bayesian_search,
        )

    @staticmethod
    def _parse_param_value(
        param_type: Optional[str], value: Any
    ) -> Tuple[ParameterType, ParameterValue]:
        if param_type is None:
            param_type = "float"
        param_type = param_type.lower().strip()

        parsed_type = ParameterType.FLOAT
        if param_type == "string":
            parsed_type = ParameterType.STRING
        elif param_type == "integer":
            parsed_type = ParameterType.INT

        if parsed_type == ParameterType.FLOAT:
            if not isinstance(value, float) and not isinstance(value, int):
                raise Exception(f"Value {value} is not of type float")
            parsed_value = ParameterValue().with_float(float(value))
        elif parsed_type == ParameterType.INT:
            if not isinstance(value, float) and not isinstance(value, int):
                raise Exception(f"Value {value} is not of type int")
            parsed_value = ParameterValue().with_int(int(value))
        elif parsed_type == ParameterType.STRING:
            if not isinstance(value, str):
                raise Exception(f"Value {value} is not of type string")
            parsed_value = ParameterValue().with_string(value)
        else:
            raise LayerClientException("Unsupported parameter type")

        return parsed_type, parsed_value

    def _parse_manual_strategy(self, payload: Dict[str, Any]) -> ManualSearch:
        parameters_manual: List[List[TypedParameter]] = []
        for values_map in payload["tuning"]:
            ls = []
            for param_name, param_body in values_map.items():
                parsed_type, parsed_value = self._parse_param_value(
                    param_body.get("type", None), param_body["value"]
                )
                ls.append(
                    TypedParameter(
                        name=param_name,
                        value=parsed_value,
                        type=parsed_type,
                    )
                )
            parameters_manual.append(ls)
        return ManualSearch(parameters=parameters_manual)

    def _parse_random_strategy(self, payload: Dict[str, Any]) -> RandomSearch:
        parameters_random: List[ParameterRange] = []
        parameters_categorical: List[ParameterCategoricalRange] = []
        for param_name, param_data in payload["tuning"].items():
            if "values" in param_data:
                param_combination = []
                for param_val in param_data["values"]:
                    parsed_type, parsed_value = self._parse_param_value(
                        param_data.get("type", None), param_val
                    )
                    param_combination.append(parsed_value)
                parsed_type, _ = self._parse_param_value(
                    param_data.get("type", None), param_data["values"][0]
                )
                parameters_categorical.append(
                    ParameterCategoricalRange(
                        name=param_name,
                        values=param_combination,
                        type=parsed_type,
                    )
                )
            elif "min" in param_data and "max" in param_data:
                parsed_type, parsed_min = self._parse_param_value(
                    param_data.get("type", None), param_data["min"]
                )
                parsed_type, parsed_max = self._parse_param_value(
                    param_data.get("type", None), param_data["max"]
                )
                parameters_random.append(
                    ParameterRange(
                        name=param_name,
                        min=parsed_min,
                        max=parsed_max,
                        type=parsed_type,
                    )
                )
            else:
                raise Exception(f"Illegal formatting of parameter {param_name}")
        return RandomSearch(
            max_jobs=payload["max_trials"]
            if "max_trials" in payload
            else payload["max_parallel_trials"],
            parameters=parameters_random,
            parameters_categorical=parameters_categorical,
        )

    def _parse_grid_strategy(self, payload: Dict[str, Any]) -> GridSearch:
        parameters_grid: List[ParameterStepRange] = []
        for param_name, param_data in payload["tuning"].items():
            parsed_type, parsed_min = self._parse_param_value(
                param_data.get("type", None), param_data["min"]
            )
            parsed_type, parsed_max = self._parse_param_value(
                param_data.get("type", None), param_data["max"]
            )
            parsed_type, parsed_step = self._parse_param_value(
                param_data.get("type", None), param_data["step"]
            )
            parameters_grid.append(
                ParameterStepRange(
                    name=param_name,
                    min=parsed_min,
                    max=parsed_max,
                    step=parsed_step,
                    type=parsed_type,
                )
            )
        return GridSearch(parameters=parameters_grid)

    def _parse_bayesian_strategy(self, payload: Dict[str, Any]) -> BayesianSearch:
        parameters_bayesian: List[ParameterRange] = []
        for param_name, param_data in payload["tuning"].items():
            parsed_type, parsed_min = self._parse_param_value(
                param_data.get("type", None), param_data["min"]
            )
            parsed_type, parsed_max = self._parse_param_value(
                param_data.get("type", None), param_data["max"]
            )
            parameters_bayesian.append(
                ParameterRange(
                    name=param_name,
                    min=parsed_min,
                    max=parsed_max,
                    type=parsed_type,
                )
            )
        return BayesianSearch(
            max_jobs=payload["max_trials"]
            if "max_trials" in payload
            else payload["max_parallel_trials"],
            parameters=parameters_bayesian,
        )
