import datetime
import hashlib
import itertools
import logging
import os
import platform
import shutil
import sys
import tempfile
import uuid
import warnings
from collections import defaultdict, deque, namedtuple
from dataclasses import dataclass, field, replace
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import pkg_resources
import yaml
from rich.text import Text

from layer import __version__
from layer.api.entity.operations_pb2 import (
    DatasetBuildOperation,
    ExecutionPlan,
    FeatureSetBuildOperation,
    HyperparameterTuningOperation,
    ModelTrainOperation,
    Operation,
    ParallelOperation,
    SequentialOperation,
)
from layer.api.ids_pb2 import HyperparameterTuningId, ModelVersionId
from layer.client import (
    DerivedDataset,
    Featureset,
    LayerClientException,
    Model,
    RawDataset,
)
from layer.common import LayerClient
from layer.exceptions.exceptions import (
    InvalidEntityTypeYamlDefinitionError,
    NoProjectFoundException,
    ProjectCircularDependenciesException,
    ProjectDependencyNotFoundException,
    ProjectException,
    ProjectInvalidYamlException,
)
from layer.parsers.derived_dataset_parser import DerivedDatasetDefinitionParser
from layer.parsers.featureset_parser import FeaturesetDefinitionParser
from layer.parsers.model_parser import ModelDefinitionParser
from layer.parsers.project_parser import ProjectYamlDefinitionParser
from layer.parsers.raw_dataset_parser import RawDatasetDefinitionParser
from layer.projects.asset import AssetType, BaseAsset
from layer.projects.constants import CWD_PATH
from layer.projects.entity import EntityType
from layer.user.account import Account


if TYPE_CHECKING:
    from networkx import DiGraph  # type: ignore

logger = logging.getLogger()
LeveledNode = namedtuple("LeveledNode", ["node", "level"])


@dataclass(frozen=True)
class PlanNode:
    type: Type[BaseAsset]
    name: str
    entity: BaseAsset


@dataclass(frozen=True)
class Asset:
    type: AssetType
    name: str


@dataclass(frozen=True)
class ResourcePath:
    # Local file system path of the resource (file or dir), relative to the project dir.
    # Examples: data/test.csv, users.parquet
    path: str

    def local_relative_paths(self) -> Iterator[str]:
        """
        Map path to the absolute file system paths, checking if paths exist.
        Includes single files and files in each resource directory.

        :return: iterator of absolute file paths.
        """
        file_path = os.path.relpath(self.path)
        if not os.path.exists(file_path):
            raise FileNotFoundError(
                f"resource file or directory: {self.path} in {os.getcwd()}"
            )
        if os.path.isfile(file_path):
            yield file_path
        if os.path.isdir(file_path):
            for root, _, files in os.walk(file_path):
                for f in files:
                    dir_file_path = os.path.join(root, f)
                    yield os.path.relpath(dir_file_path)


@dataclass(frozen=True)
class Function:
    name: str
    asset: Asset
    resource_paths: Set[ResourcePath] = field(default_factory=set)


@dataclass(frozen=True)
class Project:
    name: str
    raw_datasets: Sequence[RawDataset] = field(default_factory=list)
    derived_datasets: Sequence[DerivedDataset] = field(default_factory=list)
    featuresets: Sequence[Featureset] = field(default_factory=list)
    models: Sequence[Model] = field(default_factory=list)
    path: Path = field(compare=False, default=Path())
    project_files_hash: str = ""
    readme: str = ""
    account: Optional[Account] = None
    _id: Optional[uuid.UUID] = None
    functions: Sequence[Function] = field(default_factory=list)

    @property
    def id(self) -> uuid.UUID:
        if self._id is None:
            raise ProjectException("project has no id defined")
        return self._id

    def with_name(self, name: str) -> "Project":
        """
        :return: A new object that has a new name but all other fields are the same.
        """
        return replace(self, name=name)

    def with_id(self, project_id: uuid.UUID) -> "Project":
        """
        :return: A new object that has a new id but all other fields are the same.
        """
        return replace(self, _id=project_id)

    def with_account(self, account: Account) -> "Project":
        return replace(self, account=account)

    def with_raw_datasets(self, raw_datasets: Iterable[RawDataset]) -> "Project":
        return replace(self, raw_datasets=list(raw_datasets))

    def with_derived_datasets(
        self, derived_datasets: Iterable[DerivedDataset]
    ) -> "Project":
        return replace(self, derived_datasets=list(derived_datasets))

    def with_featuresets(self, featuresets: Iterable[Featureset]) -> "Project":
        return replace(self, featuresets=list(featuresets))

    def with_models(self, models: Iterable[Model]) -> "Project":
        return replace(self, models=list(models))

    def with_path(self, path: Path) -> "Project":
        return replace(self, path=path)

    def with_files_hash(self, new_hash: str) -> "Project":
        return replace(self, project_files_hash=new_hash)

    def with_readme(self, readme: Optional[str]) -> "Project":
        return replace(self, readme=readme)

    def with_functions(self, functions: Sequence[Function]) -> "Project":
        return replace(self, functions=functions)

    def build_execution_plan(
        self,
        models_metadata: Dict[str, ModelVersionId],
        hyperparameter_tuning_metadata: Dict[str, HyperparameterTuningId],
    ) -> ExecutionPlan:
        graph = self._build_dependency_graph()
        plan = self.topological_sort_grouping(graph)
        operations = []
        for _level, ops in plan.items():
            if len(ops) > 1:
                derived_dataset_build_operations = []
                feature_set_build_operations = []
                model_train_operations = []
                hyperparameter_tuning_operations = []
                for operation in ops:
                    dependencies = [d.path for d in operation.entity.dependencies]
                    if operation.type == Model:
                        if operation.name in hyperparameter_tuning_metadata:
                            hpt_operation = HyperparameterTuningOperation(
                                hyperparameter_tuning_id=hyperparameter_tuning_metadata[
                                    operation.name
                                ],
                                dependency=dependencies,
                            )
                            hyperparameter_tuning_operations.append(hpt_operation)
                        else:
                            model_train_operation = ModelTrainOperation(
                                model_version_id=models_metadata[operation.name],
                                dependency=dependencies,
                            )
                            model_train_operations.append(model_train_operation)
                    elif operation.type == Featureset:
                        feature_set_operation = FeatureSetBuildOperation(
                            feature_set_name=operation.name,
                            dependency=dependencies,
                        )
                        feature_set_build_operations.append(feature_set_operation)
                    elif issubclass(operation.type, DerivedDataset):
                        dataset_build_operation = DatasetBuildOperation(
                            dataset_name=operation.name,
                            dependency=dependencies,
                        )
                        derived_dataset_build_operations.append(dataset_build_operation)
                    else:
                        raise LayerClientException(
                            f"Unknown operation type. {operation}"
                        )
                operations.append(
                    Operation(
                        parallel=ParallelOperation(
                            dataset_build=derived_dataset_build_operations,
                            feature_set_build=feature_set_build_operations,
                            model_train=model_train_operations,
                            hyperparameter_tuning=hyperparameter_tuning_operations,
                        )
                    )
                )
            else:
                (operation,) = ops
                dependencies = [d.path for d in operation.entity.dependencies]
                if operation.type == Model:
                    if operation.name in hyperparameter_tuning_metadata:
                        hpt_operation = HyperparameterTuningOperation(
                            hyperparameter_tuning_id=hyperparameter_tuning_metadata[
                                operation.name
                            ],
                            dependency=dependencies,
                        )
                        operations.append(
                            Operation(
                                sequential=SequentialOperation(
                                    hyperparameter_tuning=hpt_operation
                                )
                            )
                        )
                    else:
                        model_train_operation = ModelTrainOperation(
                            model_version_id=models_metadata[operation.name],
                            dependency=dependencies,
                        )
                        operations.append(
                            Operation(
                                sequential=SequentialOperation(
                                    model_train=model_train_operation
                                )
                            )
                        )
                elif operation.type == Featureset:
                    feature_set_operation = FeatureSetBuildOperation(
                        feature_set_name=operation.name,
                        dependency=dependencies,
                    )
                    operations.append(
                        Operation(
                            sequential=SequentialOperation(
                                feature_set_build=feature_set_operation
                            )
                        )
                    )
                elif issubclass(operation.type, DerivedDataset):
                    dataset_build_operation = DatasetBuildOperation(
                        dataset_name=operation.name,
                        dependency=dependencies,
                    )
                    operations.append(
                        Operation(
                            sequential=SequentialOperation(
                                dataset_build=dataset_build_operation
                            )
                        )
                    )
                else:
                    raise LayerClientException(f"Unknown operation type. {operation}")
        execution_plan = ExecutionPlan(operations=operations)
        return execution_plan

    @staticmethod
    def topological_sort_grouping(
        graph: "DiGraph",
    ) -> DefaultDict[int, List[PlanNode]]:
        _graph = graph.copy()
        res = defaultdict(list)
        level = 0
        while _graph:
            zero_in_degree = [
                vertex for vertex, degree in _graph.in_degree() if degree == 0
            ]
            res[level] = [
                PlanNode(vertex[0], vertex[1], _graph.nodes[vertex]["entity"])
                for vertex in zero_in_degree
            ]
            _graph.remove_nodes_from(zero_in_degree)
            level = level + 1
        return res

    def _build_dependency_graph(self) -> "DiGraph":
        from networkx import DiGraph, is_directed_acyclic_graph

        graph = DiGraph()
        entities: Sequence[BaseAsset] = [
            *self.derived_datasets,
            *self.featuresets,
            *self.models,
        ]

        for entity in entities:
            name = self._get_entity_id(entity)
            graph.add_node(name, entity=entity)
        for entity in entities:
            entity_id = self._get_entity_id(entity)
            for dependency in entity.dependencies:
                dependency_entity_id = self._get_entity_id(dependency)
                # we add connections only to other entities to build
                for node_entity_id in graph.nodes:
                    if self._is_same_entity(dependency_entity_id, node_entity_id):
                        graph.add_edge(node_entity_id, entity_id)

        if not is_directed_acyclic_graph(graph):
            cycles: List[List[BaseAsset]] = self.find_cycles(graph.reverse())
            stringified_paths = [
                self._stringify_entity_cycle(cycle) for cycle in cycles
            ]
            stringified_paths.sort()  # Ensure stability across different runs
            raise ProjectCircularDependenciesException(stringified_paths)
        return graph

    @staticmethod
    def _is_same_entity(
        dependency_entity_id: Tuple[Type[BaseAsset], str],
        node_entity_id: Tuple[Type[BaseAsset], str],
    ) -> bool:
        (dependency_type, dependency_name) = dependency_entity_id
        (node_type, node_name) = node_entity_id
        # graph is build from concrete classes, while dependencies are specified by users using abstract
        # classes. we need to account for that
        return issubclass(node_type, dependency_type) and node_name == dependency_name

    @staticmethod
    def find_cycles(graph: "DiGraph") -> List[List[BaseAsset]]:
        from networkx import get_node_attributes, simple_cycles

        cycle_paths: List[List[BaseAsset]] = []
        entities_map = get_node_attributes(graph, "entity")
        for cycle in simple_cycles(graph):
            cycle_path: List[BaseAsset] = [entities_map[node] for node in cycle]
            cycle_paths.append(cycle_path)
        return cycle_paths

    @staticmethod
    def _stringify_entity_cycle(entity_cycle_path: List[BaseAsset]) -> str:
        def rotate_list(entity_cycle_path: List[str]) -> List[str]:
            smallest_idx = 0
            for i in range(1, len(entity_cycle_path)):
                if entity_cycle_path[i] < entity_cycle_path[smallest_idx]:
                    smallest_idx = i
            rotated = deque(entity_cycle_path)
            rotated.rotate(-smallest_idx)
            return list(rotated)

        stringified = rotate_list(
            [str(entity) for entity in entity_cycle_path]
        )  # Ensure stability within a cycle path across runs
        stringified.append(stringified[0])  # Add dependency to first node in cycle
        return " -> ".join(stringified)

    @staticmethod
    def _get_entity_id(entity: BaseAsset) -> Tuple[Type[BaseAsset], str]:
        return type(entity), entity.name

    @staticmethod
    def _create_not_found_exception(
        entity_id: Tuple[Type[BaseAsset], str]
    ) -> Exception:
        type_, name = entity_id
        return ProjectDependencyNotFoundException(
            f"{type_.__name__.lower().capitalize()} {name!r} not found",
            "Declare the dependency in your project",
        )

    def check_entity_dependencies(self) -> None:
        self._build_dependency_graph()

    def drop_independent_entities(
        self, type_: "EntityType", name: "str", *, keep_dependencies: bool = True
    ) -> "Project":
        from networkx import NodeNotFound, shortest_path

        target_entity_id = type_.get_factory(), name
        graph = self._build_dependency_graph()
        try:
            # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
            entity_ids: Set[Tuple[Type[BaseAsset], str]] = set.union(
                set(), *shortest_path(graph, target=target_entity_id).values()
            )
        except NodeNotFound:
            raise self._create_not_found_exception(target_entity_id)
        if not keep_dependencies:
            entity_ids = {target_entity_id}
        raw_datasets = [e for e in self.raw_datasets if keep_dependencies]
        derived_datasets = [
            e if keep_dependencies else e.drop_dependencies()
            for e in self.derived_datasets
            if self._get_entity_id(e) in entity_ids
        ]
        featuresets = [
            e if keep_dependencies else e.drop_dependencies()
            for e in self.featuresets
            if self._get_entity_id(e) in entity_ids
        ]
        models = [
            e if keep_dependencies else e.drop_dependencies()
            for e in self.models
            if self._get_entity_id(e) in entity_ids
        ]
        return (
            self.with_raw_datasets(raw_datasets)
            .with_derived_datasets(derived_datasets)
            .with_featuresets(featuresets)
            .with_models(models)
        )


@dataclass(frozen=True)
class ApplyResult:
    execution_plan: ExecutionPlan
    models_metadata: Dict[str, ModelVersionId] = field(default_factory=dict)
    hyperparameter_tuning_metadata: Dict[str, HyperparameterTuningId] = field(
        default_factory=dict
    )

    def with_models_metadata(
        self,
        models_metadata: Dict[str, ModelVersionId],
        hyperparameter_tuning_metadata: Dict[str, HyperparameterTuningId],
        plan: ExecutionPlan,
    ) -> "ApplyResult":
        return replace(
            self,
            models_metadata=dict(models_metadata),
            hyperparameter_tuning_metadata=dict(hyperparameter_tuning_metadata),
            execution_plan=plan,
        )


class ProjectPathFinder:
    _yaml_extensions = {".yml", ".yaml"}

    def find_project_configuration_path(self, root: Path) -> Path:
        search_path = Path(root / ".layer")
        project_path = None
        for path in search_path.glob("project.*"):
            if "".join(path.suffixes) not in self._yaml_extensions:
                continue
            if project_path is None:
                project_path = path
            else:
                raise ProjectException(
                    f"Duplicate project configurations found in {path}"
                )
        if project_path is None:
            raise NoProjectFoundException(root)
        return project_path

    @staticmethod
    def find_all_layer_entity_paths(
        root: Path,
    ) -> Iterable[Path]:
        all_paths = set(root.rglob("*.yml")).union(set(root.rglob("*.yaml")))
        # entities shouldn't be defined under .layer directory
        all_paths = all_paths - set(root.rglob(".layer/*"))
        return all_paths

    def find_root_path(self, path: Path) -> Optional[Path]:
        for root in [path] + list(path.parents):
            try:
                self.find_project_configuration_path(root)
                return root
            except NoProjectFoundException:
                continue
        return None

    def find_root_path_or_raise(self, path: Path) -> Path:
        root = self.find_root_path(path)
        if root is not None:
            return root
        raise NoProjectFoundException(path)


class ProjectManager:
    def __init__(
        self,
        path: Path = CWD_PATH,
    ) -> None:
        self._path = path
        self._path_finder = ProjectPathFinder()
        self._project_entity_loader = ProjectLoader()

    def clone(
        self,
        repo_url: str,
        path_to_save: str = "",
        branch: str = "main",
        *,
        git_clone_func: Optional[Callable[..., None]] = None,
    ) -> None:
        from git import Repo  # type: ignore

        if git_clone_func is None:
            git_clone_func = Repo.clone_from
        with tempfile.TemporaryDirectory() as tmpdirname:
            path_to_save = (
                path_to_save
                if path_to_save
                else self._extract_repo_name_from_url(repo_url)
            )
            destination_local_path_to_save = self._path / path_to_save
            self._check_directory_availability(destination_local_path_to_save)

            # by depth=1, we are ensuring a shallow copy.
            # https://github.com/gitpython-developers/GitPython/issues/269#issuecomment-77743473
            git_clone_func(repo_url, tmpdirname, branch=branch, depth=1)

            shutil.rmtree(Path(tmpdirname) / ".git")
            shutil.copytree(tmpdirname, destination_local_path_to_save)
            print(f"Cloned into {destination_local_path_to_save!s}")

    @staticmethod
    def _extract_repo_name_from_url(repo_url: str) -> str:
        repo_tokens = repo_url.split("/")
        repo_name = repo_tokens[-1]
        if repo_name.endswith(".git"):
            repo_name = repo_name[:-4]
        return repo_name

    def init(
        self,
        project_name: Optional[str] = None,
        directory_path: Optional[Path] = None,
        safe_init: bool = True,
        src_template_path: Optional[Path] = None,
    ) -> None:
        target_directory_path = directory_path if directory_path else self._path
        if safe_init:
            self._check_safe_init(target_directory_path)

        project_name = (
            project_name if project_name else os.path.basename(target_directory_path)
        )

        if src_template_path:
            if not src_template_path.is_dir():
                raise ProjectException(
                    f"Invalid project template at: {src_template_path!r}"
                )
            copy_tree(str(src_template_path), str(target_directory_path))
        else:
            self._init_project_yaml(target_directory_path, project_name=project_name)

        print(f"Initialized project into {target_directory_path!s}")

    @staticmethod
    def _init_project_yaml(target_directory_path: Path, project_name: str) -> None:
        project_yaml_content = f"""# Project configuration file
apiVersion: 1

# Name of your project
name: {project_name}
"""
        layer_config_directory_path = target_directory_path / ".layer"
        layer_config_directory_path.mkdir(parents=True, exist_ok=True)
        with open(layer_config_directory_path / "project.yaml", "w") as f:
            f.write(project_yaml_content)

    def _check_safe_init(self, target_directory_path: Path) -> None:
        self._check_directory_availability(target_directory_path)

        project_root = self._path_finder.find_root_path(target_directory_path)

        if project_root:
            raise ProjectException(
                f"Your project is already initialized at {project_root!s}"
            )

    @staticmethod
    def _check_directory_availability(directory_path: Path) -> None:
        if directory_path.exists() and (
            not directory_path.is_dir()
            or next(directory_path.iterdir(), None) is not None
        ):
            raise ProjectException(
                f"Destination path {directory_path!s} already exists and is not an empty directory"
            )

    def load(self) -> Project:
        root = self._path_finder.find_root_path_or_raise(self._path)
        print(
            Text.from_markup(":file_folder:").append(
                f" Loading the project under {root}"
            )
        )
        project_path = self._path_finder.find_project_configuration_path(root)
        project = self._project_entity_loader.load_project(project_path)
        all_paths = ProjectPathFinder.find_all_layer_entity_paths(root)
        project_files_hash = self._project_entity_loader.compute_project_files_digest(
            itertools.chain(all_paths, [project_path])
        )
        (
            raw_datasets,
            derived_datasets,
            featuresets,
            models,
        ) = self._project_entity_loader.load_entities(
            all_paths, project_name=project.name
        )
        readme = self._project_entity_loader.load_project_readme(root)
        print(
            Text.from_markup(":file_folder:").append(
                f" Successfully loaded project with name: {project.name}"
            )
        )
        return (
            project.with_files_hash(project_files_hash)
            .with_raw_datasets(raw_datasets)
            .with_derived_datasets(derived_datasets)
            .with_featuresets(featuresets)
            .with_models(models)
            .with_readme(readme)
            .with_path(root)
        )

    def diagnose(self) -> None:
        print(f"Local diagnostics for Layer {__version__}")

        print(f"OS: {platform.system()} {platform.release()}")
        print(f"Platform: {platform.uname()}")
        print(f"Interpreter path: {sys.executable}")
        print(f"Interpreter version: {sys.version}")
        tz_string = datetime.datetime.now(datetime.timezone.utc).astimezone().tzname()
        print(f"Timezone: {tz_string}")
        env_var_keys = " ".join(dict(os.environ).keys())
        print(f"Environment variable keys: {env_var_keys}")

        path_finder = ProjectPathFinder()
        try:
            path_finder.find_project_configuration_path(Path.cwd())
            print(
                "Layer project found in the current directory. Printing the directory structure:"
            )
            for line in self._tree(Path.cwd()):
                print(line)
        except NoProjectFoundException:
            print("The current working directory does not have a Layer project.")

        global_layer_config = Path.home() / ".layer"
        if global_layer_config.exists():
            is_empty = not os.listdir(global_layer_config)
            print(
                f"{global_layer_config} is found"
                + (". Contents:" if not is_empty else " but it is empty!")
            )
            for line in self._tree(global_layer_config):
                print(line)

        installed_packages = sorted(
            [
                f"{package.project_name} {package.version}"
                for package in pkg_resources.working_set
            ]
        )
        installed_packages_str = " ".join(installed_packages)
        print(f"Installed packages: {installed_packages_str}")

    # https://stackoverflow.com/a/59109706/126199
    def _tree(self, dir_path: Path, prefix: str = "") -> Iterator[str]:
        """A recursive generator, given a directory Path object
        will yield a visual tree structure line by line
        with each line prefixed by the same characters
        """
        # prefix components:
        space = "    "
        branch = "│   "
        # pointers:
        tee = "├── "
        last = "└── "

        contents = list(dir_path.iterdir())
        # contents each get pointers that are ├── with a final └── :
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            yield prefix + pointer + path.name
            if path.is_dir():  # extend the prefix and recurse:
                extension = branch if pointer == tee else space
                # i.e. space because last, └── , above so no more |
                yield from self._tree(path, prefix=prefix + extension)


class ProjectLoader:
    @staticmethod
    def load_project(path: Path) -> Project:
        parser = ProjectYamlDefinitionParser(path)
        parser.validate()
        project_definition = parser.parse()
        return Project(name=project_definition.name)

    @staticmethod
    def compute_project_files_digest(project_files: Iterable[Path]) -> str:
        project_files_hash = hashlib.sha256()
        for f_path in project_files:
            with open(f_path, "rb") as project_file:
                file_bytes = project_file.read()
                project_files_hash.update(file_bytes)
        return project_files_hash.hexdigest()

    @staticmethod
    def load_entities(
        all_paths: Iterable[Path],
        project_name: str,
    ) -> Tuple[List[RawDataset], List[DerivedDataset], List[Featureset], List[Model]]:
        all_entities_by_type: Dict[
            EntityType, Dict[str, Union[RawDataset, DerivedDataset, Featureset, Model]]
        ] = {
            EntityType.RAW_DATASET: {},
            EntityType.DERIVED_DATASET: {},
            EntityType.MODEL: {},
            EntityType.FEATURESET: {},
        }
        for path in all_paths:
            with open(path, "r") as f:
                try:
                    for payload in yaml.safe_load_all(
                        f
                    ):  # it's possible to define multiple entities in one yaml file
                        (
                            entity_type,
                            entity_name,
                            entity,
                        ) = ProjectLoader._load_single_layer_entity(
                            payload, path, project_name, all_entities_by_type
                        )
                        all_entities_by_type[entity_type][entity_name] = entity
                except yaml.YAMLError as e:
                    raise ProjectInvalidYamlException(path, e)
        if (
            not all_entities_by_type[EntityType.MODEL].keys()
            and not all_entities_by_type[EntityType.FEATURESET].keys()
            and not all_entities_by_type[EntityType.RAW_DATASET].keys()
            and not all_entities_by_type[EntityType.DERIVED_DATASET].keys()
        ):
            raise ProjectException(
                "You should define at least one of raw dataset, derived dataset, featureset or model."
            )
        raw_datasets = cast(
            List[RawDataset],
            list(all_entities_by_type[EntityType.RAW_DATASET].values()),
        )
        derived_datasets = cast(
            List[DerivedDataset],
            list(all_entities_by_type[EntityType.DERIVED_DATASET].values()),
        )
        featuresets = cast(
            List[Featureset],
            list(all_entities_by_type[EntityType.FEATURESET].values()),
        )
        models = cast(
            List[Model],
            list(all_entities_by_type[EntityType.MODEL].values()),
        )
        return raw_datasets, derived_datasets, featuresets, models

    @staticmethod
    def _load_single_layer_entity(
        payload: Dict[str, Any],
        path: Path,
        project_name: str,
        all_entities_by_type: Dict[
            EntityType, Dict[str, Union[RawDataset, DerivedDataset, Featureset, Model]]
        ],
    ) -> Tuple[EntityType, str, Union[RawDataset, DerivedDataset, Featureset, Model]]:
        entity_type = ProjectLoader._get_entity_type_from_payload(payload, path)
        entity = ProjectLoader._parse_entity(entity_type, payload, path, project_name)
        if entity.name in all_entities_by_type[entity_type]:
            raise ProjectException(
                f"Duplicate {entity_type} configurations with name '{entity.name}' found in {path}"
            )
        return entity_type, entity.name, entity

    @staticmethod
    def load_project_readme(path: Path) -> Optional[str]:
        for subp in path.glob("*"):
            if subp.name.lower() == "readme.md":
                with open(subp, "r") as f:
                    readme = f.read()
                    # restrict length of text we send to backend
                    if len(readme) > 25_000:
                        warnings.warn(
                            "Your README.md will be truncated to 25000 characters",
                        )
                        readme = readme[:25_000]
                    return readme
        return None

    @staticmethod
    def _parse_entity(
        entity_type: EntityType,
        payload: Dict[str, Any],
        path: Path,
        project_name: str,
    ) -> Union[RawDataset, DerivedDataset, Featureset, Model]:
        if entity_type == EntityType.RAW_DATASET:
            parser = RawDatasetDefinitionParser(
                payload, path, project_name=project_name
            )
        elif entity_type == EntityType.DERIVED_DATASET:
            parser = DerivedDatasetDefinitionParser(payload, path, project_name=project_name)  # type: ignore
        elif entity_type == EntityType.FEATURESET:
            parser = FeaturesetDefinitionParser(payload, path, project_name=project_name)  # type: ignore
        elif entity_type == EntityType.MODEL:
            parser = ModelDefinitionParser(payload, path, project_name=project_name)  # type: ignore
        else:
            raise InvalidEntityTypeYamlDefinitionError(path, entity_type)
        parser.validate()
        return parser.parse()

    @staticmethod
    def _get_entity_type_from_payload(
        payload: Dict[str, Any], path: Path
    ) -> EntityType:
        entity_type_str = payload.get("type")
        if entity_type_str is None:
            raise ProjectException(
                f"Layer entity defined in {path} is missing required field 'type'"
            )
        try:
            entity_type = EntityType(entity_type_str)
        except ValueError:
            raise InvalidEntityTypeYamlDefinitionError(path, entity_type_str)
        return entity_type


def get_or_create_remote_project(client: LayerClient, project: Project) -> Project:
    project_id_with_org_id = client.project_service_client.get_project_id_and_org_id(
        project.name
    )
    if project_id_with_org_id.project_id is None:
        project_id_with_org_id = client.project_service_client.create_project(
            project.name
        )
    assert project_id_with_org_id.project_id is not None
    assert project_id_with_org_id.account_id is not None
    account_name = client.account.get_account_name_by_id(
        project_id_with_org_id.account_id
    )
    account = Account(id=project_id_with_org_id.account_id, name=account_name)
    return project.with_id(project_id_with_org_id.project_id).with_account(
        account=account
    )
