from logging import Logger
from typing import Any, Callable, Dict, Optional

import pandas as pd

import layer
from layer import Context, Dataset, Featureset, Model
from layer.annotation_processor import parse_entity_dependencies


def non_optional_current_project_name() -> str:
    project_name = layer.current_project_name()
    assert project_name is not None
    return project_name


def inject_annotated_dependencies(
    func: Callable[..., Any],
    context: Context,
    logger: Logger,
    features_df_factory: Optional[Callable[..., pd.DataFrame]] = None,
    featureset_factory: Callable[..., Featureset] = layer.get_featureset,
    dataset_factory: Callable[..., Dataset] = layer.get_dataset,
    model_factory: Callable[..., Model] = layer.get_model,
) -> Dict[str, Any]:
    if features_df_factory is None:

        def df_factory(feature_set_name: str, feature_names: Any) -> pd.DataFrame:
            feature_set = featureset_factory(feature_set_name, feature_names)
            return feature_set.to_pandas()

        features_df_factory = df_factory

    dependency_arguments = parse_entity_dependencies(func)
    logger.debug(f"Entity dependencies: {dependency_arguments.__dict__}")
    injected_params: Dict[str, Any] = {}
    for arg_name in dependency_arguments.featuresets:
        featureset = dependency_arguments.featuresets[arg_name]
        if not featureset.project_name:
            featureset = featureset.with_project_name(
                non_optional_current_project_name()
            )
        logger.debug(
            f"Injecting {featureset.path} featureset with individual features {featureset.feature_names}"
        )
        if len(featureset.feature_names) == 0:
            injected_params[arg_name] = featureset_factory(name=featureset.path)
        else:
            pandas_df_factory = new_pandas_df_factory(
                features_df_factory, featureset.path, featureset.feature_names
            )

            injected_params[arg_name] = Featureset(
                asset_path=featureset.path,
                feature_names=featureset.feature_names,
                _pandas_df_factory=pandas_df_factory,
            )

    for arg_name in dependency_arguments.raw_datasets:
        raw_dataset = dependency_arguments.raw_datasets[arg_name]
        if not raw_dataset.project_name:
            raw_dataset = raw_dataset.with_project_name(
                non_optional_current_project_name()
            )
        logger.debug(f"Injecting {raw_dataset.path} raw dataset")
        injected_params[arg_name] = dataset_factory(name=raw_dataset.path)
    for arg_name in dependency_arguments.derived_datasets:
        derived_dataset = dependency_arguments.derived_datasets[arg_name]
        if not derived_dataset.project_name:
            derived_dataset = derived_dataset.with_project_name(
                non_optional_current_project_name()
            )
        logger.debug(f"Injecting {derived_dataset.path} derived dataset")
        injected_params[arg_name] = dataset_factory(name=derived_dataset.path)
    for arg_name in dependency_arguments.models:
        model = dependency_arguments.models[arg_name]
        if not model.project_name:
            model = model.with_project_name(non_optional_current_project_name())
        logger.debug(f"Injecting {model.path} model")
        injected_params[arg_name] = model_factory(name=model.path)

    if dependency_arguments.context is not None:
        injected_params[dependency_arguments.context] = context
    if dependency_arguments.train is not None:
        injected_params[dependency_arguments.train] = context.train()
    return injected_params


def new_pandas_df_factory(
    features_df_factory: Callable[..., pd.DataFrame],
    featureset_name: str,
    feature_names: Any,
) -> Callable[[], pd.DataFrame]:
    def pandas_df_factory() -> pd.DataFrame:
        return features_df_factory(featureset_name, feature_names)

    return pandas_df_factory
